#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains support for Shadow Registers Tool."""
from .shadowregs import ShadowRegisters, enable_debug
