#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""SPSDK version.

Having the version in a separate file makes it easier to share it with setup.py
"""

__version__ = "1.11.0"
